@@ -30,13 +30,14 @@ package org.apache.hc.core5.http.impl.nio;
 import java.io.IOException;
 import java.net.SocketAddress;
 import java.nio.ByteBuffer;
-import java.nio.channels.ClosedChannelException;
 import java.nio.channels.ReadableByteChannel;
 import java.nio.channels.SelectionKey;
 import java.nio.channels.WritableByteChannel;
 import java.util.List;
 import java.util.concurrent.atomic.AtomicInteger;
+import java.util.concurrent.locks.ReentrantLock;
 
+import javax.net.ssl.SSLHandshakeException;
 import javax.net.ssl.SSLSession;
 
 import org.apache.hc.core5.http.ConnectionClosedException;
@@ -116,10 +117,10 @@ abstract class AbstractHttp1StreamDuplexer<IncomingMessage extends HttpMessage,
         this.ioSession = Args.notNull(ioSession, "I/O session");
         this.http1Config = http1Config != null ? http1Config : Http1Config.DEFAULT;
         final int bufferSize = this.http1Config.getBufferSize();
-        this.inbuf = new SessionInputBufferImpl(bufferSize, bufferSize < 512 ? bufferSize : 512,
+        this.inbuf = new SessionInputBufferImpl(bufferSize, Math.min(bufferSize, 512),
                 this.http1Config.getMaxLineLength(),
                 CharCodingSupport.createDecoder(charCodingConfig));
-        this.outbuf = new SessionOutputBufferImpl(bufferSize, bufferSize < 512 ? bufferSize : 512,
+        this.outbuf = new SessionOutputBufferImpl(bufferSize, Math.min(bufferSize, 512),
                 CharCodingSupport.createEncoder(charCodingConfig));
         this.inTransportMetrics = new BasicHttpTransportMetrics();
         this.outTransportMetrics = new BasicHttpTransportMetrics();
@@ -140,6 +141,14 @@ abstract class AbstractHttp1StreamDuplexer<IncomingMessage extends HttpMessage,
         return ioSession.getId();
     }
 
+    boolean isActive() {
+        return connState == ConnectionState.ACTIVE;
+    }
+
+    boolean isShuttingDown() {
+        return connState.compareTo(ConnectionState.GRACEFUL_SHUTDOWN) >= 0;
+    }
+
     void shutdownSession(final CloseMode closeMode) {
         if (closeMode == CloseMode.GRACEFUL) {
             connState = ConnectionState.GRACEFUL_SHUTDOWN;
@@ -158,6 +167,8 @@ abstract class AbstractHttp1StreamDuplexer<IncomingMessage extends HttpMessage,
             final CloseMode closeMode;
             if (cause instanceof ConnectionClosedException) {
                 closeMode = CloseMode.GRACEFUL;
+            } else if (cause instanceof SSLHandshakeException) {
+                closeMode = CloseMode.GRACEFUL;
             } else if (cause instanceof IOException) {
                 closeMode = CloseMode.IMMEDIATE;
             } else {
@@ -254,7 +265,9 @@ abstract class AbstractHttp1StreamDuplexer<IncomingMessage extends HttpMessage,
 
     public final void onInput(final ByteBuffer src) throws HttpException, IOException {
         if (src != null) {
+            final int n = src.remaining();
             inbuf.put(src);
+            inTransportMetrics.incrementBytesTransferred(n);
         }
 
         if (connState.compareTo(ConnectionState.GRACEFUL_SHUTDOWN) >= 0 && inbuf.hasData() && inputIdle()) {
@@ -327,8 +340,7 @@ abstract class AbstractHttp1StreamDuplexer<IncomingMessage extends HttpMessage,
                     incomingMessage = null;
                     ioSession.setEvent(SelectionKey.OP_READ);
                     inputEnd();
-                }
-                if (bytesRead == 0) {
+                } else if (bytesRead == 0) {
                     break;
                 }
             }
@@ -437,7 +449,10 @@ abstract class AbstractHttp1StreamDuplexer<IncomingMessage extends HttpMessage,
             }
             outgoingMessageWriter.reset();
             if (flushMode == FlushMode.IMMEDIATE) {
-                outbuf.flush(ioSession);
+                final int bytesWritten = outbuf.flush(ioSession);
+                if (bytesWritten > 0) {
+                    outTransportMetrics.incrementBytesTransferred(bytesWritten);
+                }
             }
             ioSession.setEvent(EventMask.WRITE);
         } finally {
@@ -470,7 +485,10 @@ abstract class AbstractHttp1StreamDuplexer<IncomingMessage extends HttpMessage,
         ioSession.getLock().lock();
         try {
             if (outbuf.hasData()) {
-                outbuf.flush(ioSession);
+                final int bytesWritten = outbuf.flush(ioSession);
+                if (bytesWritten > 0) {
+                    outTransportMetrics.incrementBytesTransferred(bytesWritten);
+                }
             } else {
                 ioSession.clearEvent(SelectionKey.OP_WRITE);
             }
@@ -483,7 +501,7 @@ abstract class AbstractHttp1StreamDuplexer<IncomingMessage extends HttpMessage,
         ioSession.getLock().lock();
         try {
             if (outgoingMessage == null) {
-                throw new ClosedChannelException();
+                throw new ConnectionClosedException();
             }
             final ContentEncoder contentEncoder = outgoingMessage.getBody();
             final int bytesWritten = contentEncoder.write(src);
@@ -596,19 +614,20 @@ abstract class AbstractHttp1StreamDuplexer<IncomingMessage extends HttpMessage,
 
     static class CapacityWindow implements CapacityChannel {
         private final IOSession ioSession;
-        private final Object lock;
+        private final ReentrantLock lock;
         private int window;
         private boolean closed;
 
         CapacityWindow(final int window, final IOSession ioSession) {
             this.window = window;
             this.ioSession = ioSession;
-            this.lock = new Object();
+            this.lock = new ReentrantLock();
         }
 
         @Override
         public void update(final int increment) throws IOException {
-            synchronized (lock) {
+            lock.lock();
+            try {
                 if (closed) {
                     return;
                 }
@@ -616,6 +635,8 @@ abstract class AbstractHttp1StreamDuplexer<IncomingMessage extends HttpMessage,
                     updateWindow(increment);
                     ioSession.setEvent(SelectionKey.OP_READ);
                 }
+            } finally {
+                lock.unlock();
             }
         }
 
@@ -624,12 +645,15 @@ abstract class AbstractHttp1StreamDuplexer<IncomingMessage extends HttpMessage,
          * if this channel is closed in it.
          */
         int removeCapacity(final int delta) {
-            synchronized (lock) {
+            lock.lock();
+            try {
                 updateWindow(-delta);
                 if (window <= 0) {
                     ioSession.clearEvent(SelectionKey.OP_READ);
                 }
                 return window;
+            } finally {
+                lock.unlock();
             }
         }
 
@@ -647,8 +671,11 @@ abstract class AbstractHttp1StreamDuplexer<IncomingMessage extends HttpMessage,
          * read events outside of the context of the request the channel was created for
          */
         void close() {
-            synchronized (lock) {
+            lock.lock();
+            try {
                 closed = true;
+            } finally {
+                lock.unlock();
             }
         }
 
